# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes
from ...core import OutputType
from ...serialize import BoolField, TupleField, DictField, AnyField
from ..operands import DataFrameOperandMixin, DataFrameOperand
from ..utils import build_empty_df, build_empty_series, parse_index


class GroupByTransform(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.TRANSFORM
    _op_module_ = 'dataframe.groupby'

    _func = AnyField('func')
    _args = TupleField('args')
    _kwds = DictField('kwds')

    _call_agg = BoolField('call_agg')

    def __init__(self, func=None, args=None, kwds=None, call_agg=None, output_types=None, **kw):
        super().__init__(_func=func, _args=args, _kwds=kwds, _call_agg=call_agg,
                         _output_types=output_types, **kw)

    @property
    def func(self):
        return self._func

    @property
    def args(self):
        return getattr(self, '_args', None) or ()

    @property
    def kwds(self):
        return getattr(self, '_kwds', None) or dict()

    @property
    def call_agg(self):
        return self._call_agg

    def _infer_df_func_returns(self, in_groupby, dtypes, index):
        index_value, output_types, new_dtypes = None, None, None

        output_types = [OutputType.dataframe] \
            if in_groupby.op.output_types[0] == OutputType.dataframe_groupby else [OutputType.series]

        try:
            empty_groupby = in_groupby.op.build_mock_groupby()
            with np.errstate(all='ignore'):
                if self.call_agg:
                    infer_df = empty_groupby.agg(self.func, *self.args, **self.kwds)
                else:
                    infer_df = empty_groupby.transform(self.func, *self.args, **self.kwds)

            # todo return proper index when sort=True is implemented
            index_value = parse_index(None, in_groupby.key, self.func)

            if isinstance(infer_df, pd.DataFrame):
                output_types = [OutputType.dataframe]
                new_dtypes = new_dtypes or infer_df.dtypes
            else:
                output_types = [OutputType.series]
                new_dtypes = new_dtypes or (infer_df.name, infer_df.dtype)
        except:  # noqa: E722  # nosec
            pass

        self.output_types = output_types if not self.output_types else self.output_types
        dtypes = new_dtypes if dtypes is None else dtypes
        index_value = index_value if index is None else parse_index(index)
        return dtypes, index_value

    def __call__(self, groupby, dtypes=None, index=None):
        in_df = groupby.inputs[0]

        dtypes, index_value = self._infer_df_func_returns(groupby, dtypes, index)
        for arg, desc in zip((self.output_types, dtypes, index_value),
                             ('output_types', 'dtypes', 'index')):
            if arg is None:
                raise TypeError(f'Cannot determine {desc} by calculating with enumerate data, '
                                'please specify it as arguments')

        if self.output_types[0] == OutputType.dataframe:
            new_shape = (np.nan if self.call_agg else in_df.shape[0], len(dtypes))
            return self.new_dataframe([groupby], shape=new_shape, dtypes=dtypes, index_value=index_value,
                                      columns_value=parse_index(dtypes.index, store_data=True))
        else:
            name, dtype = dtypes
            new_shape = (np.nan,) if self.call_agg else groupby.shape
            return self.new_series([groupby], name=name, shape=new_shape, dtype=dtype,
                                   index_value=index_value)

    @classmethod
    def tile(cls, op):
        in_groupby = op.inputs[0]
        out_df = op.outputs[0]

        chunks = []
        for c in in_groupby.chunks:
            inp_chunks = [c]

            new_op = op.copy().reset_key()
            if op.output_types[0] == OutputType.dataframe:
                new_index = c.index if c.ndim == 2 else c.index + (0,)
                chunks.append(new_op.new_chunk(
                    inp_chunks, index=new_index, shape=(np.nan, len(out_df.dtypes)), dtypes=out_df.dtypes,
                    columns_value=out_df.columns_value, index_value=out_df.index_value))
            else:
                chunks.append(new_op.new_chunk(
                    inp_chunks, name=out_df.name, index=(c.index[0],), shape=(np.nan,), dtype=out_df.dtype,
                    index_value=out_df.index_value))

        new_op = op.copy().reset_key()
        kw = out_df.params.copy()
        kw['chunks'] = chunks
        if op.output_types[0] == OutputType.dataframe:
            kw['nsplits'] = ((np.nan,) * len(chunks), (len(out_df.dtypes),))
        else:
            kw['nsplits'] = ((np.nan,) * len(chunks),)
        return new_op.new_tileables([in_groupby], **kw)

    @classmethod
    def execute(cls, ctx, op):
        in_data = ctx[op.inputs[0].key]
        out_chunk = op.outputs[0]

        if not in_data:
            if op.output_types[0] == OutputType.dataframe:
                ctx[op.outputs[0].key] = build_empty_df(out_chunk.dtypes)
            else:
                ctx[op.outputs[0].key] = build_empty_series(out_chunk.dtype)
            return

        if op.call_agg:
            result = in_data.agg(op.func, *op.args, **op.kwds)
        else:
            result = in_data.transform(op.func, *op.args, **op.kwds)

        if result.ndim == 2:
            result = result.astype(op.outputs[0].dtypes, copy=False)
        else:
            result = result.astype(op.outputs[0].dtype, copy=False)
        ctx[op.outputs[0].key] = result


def groupby_transform(groupby, func, *args, dtypes=None, index=None, output_types=None, **kwargs):
    # todo this can be done with sort_index implemented
    if not groupby.op.groupby_params.get('as_index', True):
        raise NotImplementedError('transform when set_index == False is not supported')

    call_agg = kwargs.pop('_call_agg', False)
    if not call_agg and isinstance(func, (dict, list)):
        raise TypeError(f'Does not support transform with {type(func)}')

    op = GroupByTransform(func=func, args=args, kwds=kwargs, output_types=output_types,
                          call_agg=call_agg)
    return op(groupby, dtypes=dtypes, index=index)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ...serialize import Int32Field, StringField, ListField, BoolField, ValueType
from ..operands import DataFrameOperand


class DataFrameSortOperand(DataFrameOperand):
    _axis = Int32Field('axis')
    _ascending = BoolField('ascending')
    _inplace = BoolField('inplace')
    _kind = StringField('kind')
    _na_position = StringField('na_position')
    _ignore_index = BoolField('ignore_index')
    _parallel_kind = StringField('parallel_kind')
    _psrs_kinds = ListField('psrs_kinds', ValueType.string)

    def __init__(self, axis=None, ascending=None, inplace=None, kind=None,
                 na_position=None, ignore_index=None, parallel_kind=None, psrs_kinds=None, **kw):
        super(DataFrameSortOperand, self).__init__(_axis=axis, _ascending=ascending,
                                                   _inplace=inplace, _kind=kind,
                                                   _na_position=na_position,
                                                   _ignore_index=ignore_index,
                                                   _parallel_kind=parallel_kind,
                                                   _psrs_kinds=psrs_kinds, **kw)

    @property
    def axis(self):
        return self._axis

    @property
    def ascending(self):
        return self._ascending

    @property
    def inplace(self):
        return self._inplace

    @property
    def kind(self):
        return self._kind

    @property
    def na_position(self):
        return self._na_position

    @property
    def ignore_index(self):
        return self._ignore_index

    @property
    def parallel_kind(self):
        return self._parallel_kind

    @property
    def psrs_kinds(self):
        return self._psrs_kinds



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np

from ... import opcodes as OperandDef
from ...serialize import KeyField, DataTypeField, StringField
from ..array_utils import as_same_device, device
from ..operands import TensorHasInput, TensorOperandMixin
from ..utils import get_order


class TensorAstype(TensorHasInput, TensorOperandMixin):
    _op_type_ = OperandDef.ASTYPE

    _input = KeyField('input')
    _dtype = DataTypeField('dtype')
    _order = StringField('order')
    _casting = StringField('casting')

    def __init__(self, dtype=None, order=None, casting=None, sparse=False, **kw):
        super().__init__(_dtype=dtype, _order=order, _casting=casting, _sparse=sparse, **kw)

    @property
    def dtype(self):
        return self._dtype

    @property
    def order(self):
        return self._order

    @property
    def casting(self):
        return self._casting

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, tensor, order=None):
        return self.new_tensor([tensor], tensor.shape, order=order)

    @classmethod
    def tile(cls, op):
        in_tensor = op.input
        out_tensor = op.outputs[0]

        out_chunks = []
        for c in in_tensor.chunks:
            chunk_op = op.copy().reset_key()
            chunk = chunk_op.new_chunk([c], shape=c.shape, index=c.index,
                                       order=out_tensor.order)
            out_chunks.append(chunk)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, out_tensor.shape, nsplits=in_tensor.nsplits,
                                  chunks=out_chunks)

    @classmethod
    def execute(cls, ctx, op):
        chunk = op.outputs[0]
        (x,), device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True)

        with device(device_id):
            if op.sparse:
                ctx[chunk.key] = x.astype(op.dtype)
            else:
                if xp is np:
                    ctx[chunk.key] = x.astype(op.dtype, order=op.order,
                                              casting=op.casting)
                else:  # pragma: no cover
                    # cupy does not support casting
                    ctx[chunk.key] = x.astype(op.dtype, order=op.order)


def _astype(tensor, dtype, order='K', casting='unsafe', copy=True):
    """
    Copy of the tensor, cast to a specified type.

    Parameters
    ----------
    dtype : str or dtype
        Typecode or data-type to which the array is cast.
    casting : {'no', 'equiv', 'safe', 'same_kind', 'unsafe'}, optional
        Controls what kind of data casting may occur. Defaults to 'unsafe'
        for backwards compatibility.
          * 'no' means the data types should not be cast at all.
          * 'equiv' means only byte-order changes are allowed.
          * 'safe' means only casts which can preserve values are allowed.
          * 'same_kind' means only safe casts or casts within a kind,
            like float64 to float32, are allowed.
          * 'unsafe' means any data conversions may be done.
    order : {'C', 'F', 'A', 'K'}, optional
        Controls the memory layout order of the result.
        'C' means C order, 'F' means Fortran order, 'A'
        means 'F' order if all the arrays are Fortran contiguous,
        'C' order otherwise, and 'K' means as close to the
        order the array elements appear in memory as possible.
        Default is 'K'.
    copy : bool, optional
        By default, astype always returns a newly allocated array. If this
        is set to false, and the `dtype`, `order`, and `subok`
        requirements are satisfied, the input array is returned instead
        of a copy.

    Returns
    -------
    arr_t : Tensor
        Unless `copy` is False and the other conditions for returning the input
        array are satisfied (see description for `copy` input parameter), `arr_t`
        is a new tensor of the same shape as the input array, with dtype, order
        given by `dtype`, `order`.

    Notes
    -----
    astype method returns an error if the string
    dtype to cast to is not long enough in 'safe' casting mode to hold the max
    value of integer/float array that is being casted. Previously the casting
    was allowed even if the result was truncated.

    Raises
    ------
    ComplexWarning
        When casting from complex to float or int. To avoid this,
        one should use ``a.real.astype(t)``.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> x = mt.array([1, 2, 2.5])
    >>> x.execute()
    array([ 1. ,  2. ,  2.5])

    >>> x.astype(int).execute()
    array([1, 2, 2])
    """
    dtype = np.dtype(dtype)
    tensor_order = get_order(order, tensor.order)

    if tensor.dtype == dtype and tensor.order == tensor_order:
        return tensor if not copy else tensor.copy(order=order)
    elif not np.can_cast(tensor.dtype, dtype, casting=casting):
        raise TypeError(f'Cannot cast array from {tensor.dtype!r} to {dtype!r} '
                        f'according to the rule {casting}')

    op = TensorAstype(dtype=dtype, order=order, casting=casting, sparse=tensor.issparse())
    return op(tensor, order=tensor_order)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

import numpy as np

from ..core import Tensor
from ..datasource import tensor as astensor, array
from ..base.squeeze import squeeze
from ..base.where import where
from .average import average


def cov(m, y=None, rowvar=True, bias=False, ddof=None, fweights=None, aweights=None):
    """
    Estimate a covariance matrix, given data and weights.

    Covariance indicates the level to which two variables vary together.
    If we examine N-dimensional samples, :math:`X = [x_1, x_2, ... x_N]^T`,
    then the covariance matrix element :math:`C_{ij}` is the covariance of
    :math:`x_i` and :math:`x_j`. The element :math:`C_{ii}` is the variance
    of :math:`x_i`.

    See the notes for an outline of the algorithm.

    Parameters
    ----------
    m : array_like
        A 1-D or 2-D array containing multiple variables and observations.
        Each row of `m` represents a variable, and each column a single
        observation of all those variables. Also see `rowvar` below.
    y : array_like, optional
        An additional set of variables and observations. `y` has the same form
        as that of `m`.
    rowvar : bool, optional
        If `rowvar` is True (default), then each row represents a
        variable, with observations in the columns. Otherwise, the relationship
        is transposed: each column represents a variable, while the rows
        contain observations.
    bias : bool, optional
        Default normalization (False) is by ``(N - 1)``, where ``N`` is the
        number of observations given (unbiased estimate). If `bias` is True,
        then normalization is by ``N``. These values can be overridden by using
        the keyword ``ddof`` in numpy versions >= 1.5.
    ddof : int, optional
        If not ``None`` the default value implied by `bias` is overridden.
        Note that ``ddof=1`` will return the unbiased estimate, even if both
        `fweights` and `aweights` are specified, and ``ddof=0`` will return
        the simple average. See the notes for the details. The default value
        is ``None``.
    fweights : array_like, int, optional
        1-D tensor of integer freguency weights; the number of times each
        observation vector should be repeated.
    aweights : array_like, optional
        1-D tensor of observation vector weights. These relative weights are
        typically large for observations considered "important" and smaller for
        observations considered less "important". If ``ddof=0`` the array of
        weights can be used to assign probabilities to observation vectors.

    Returns
    -------
    out : Tensor
        The covariance matrix of the variables.

    See Also
    --------
    corrcoef : Normalized covariance matrix

    Notes
    -----
    Assume that the observations are in the columns of the observation
    array `m` and let ``f = fweights`` and ``a = aweights`` for brevity. The
    steps to compute the weighted covariance are as follows::

        >>> w = f * a
        >>> v1 = mt.sum(w)
        >>> v2 = mt.sum(w * a)
        >>> m -= mt.sum(m * w, axis=1, keepdims=True) / v1
        >>> cov = mt.dot(m * w, m.T) * v1 / (v1**2 - ddof * v2)

    Note that when ``a == 1``, the normalization factor
    ``v1 / (v1**2 - ddof * v2)`` goes over to ``1 / (np.sum(f) - ddof)``
    as it should.

    Examples
    --------
    Consider two variables, :math:`x_0` and :math:`x_1`, which
    correlate perfectly, but in opposite directions:

    >>> import mars.tensor as mt

    >>> x = mt.array([[0, 2], [1, 1], [2, 0]]).T
    >>> x.execute()
    array([[0, 1, 2],
           [2, 1, 0]])

    Note how :math:`x_0` increases while :math:`x_1` decreases. The covariance
    matrix shows this clearly:

    >>> mt.cov(x).execute()
    array([[ 1., -1.],
           [-1.,  1.]])

    Note that element :math:`C_{0,1}`, which shows the correlation between
    :math:`x_0` and :math:`x_1`, is negative.

    Further, note how `x` and `y` are combined:

    >>> x = [-2.1, -1,  4.3]
    >>> y = [3,  1.1,  0.12]
    >>> X = mt.stack((x, y), axis=0)
    >>> print(mt.cov(X).execute())
    [[ 11.71        -4.286     ]
     [ -4.286        2.14413333]]
    >>> print(mt.cov(x, y).execute())
    [[ 11.71        -4.286     ]
     [ -4.286        2.14413333]]
    >>> print(mt.cov(x).execute())
    11.71

    """
    from ..merge import vstack
    from ..linalg import dot

    if ddof is not None and ddof != int(ddof):
        raise ValueError(
            "ddof must be integer")

    m = astensor(m)
    if m.ndim > 2:
        raise ValueError("m has more than 2 dimensions")

    if y is None:
        dtype = np.result_type(m.dtype, np.float64)
    else:
        y = astensor(y)
        if y.ndim > 2:
            raise ValueError("y has more than 2 dimensions")
        dtype = np.result_type(m.dtype, y.dtype, np.float64)

    X = array(m, ndmin=2, dtype=dtype)
    if not rowvar and X.shape[0] != 1:
        X = X.T
    if y is not None:
        y = array(y, copy=False, ndmin=2, dtype=dtype)
        if not rowvar and y.shape[0] != 1:
            y = y.T
        X = vstack((X, y))

    if ddof is None:
        if bias == 0:
            ddof = 1
        else:
            ddof = 0

    # Get the product of frequencies and weights
    w = None
    if fweights is not None:
        fweights = astensor(fweights, dtype=np.float)
        if fweights.ndim > 1:
            raise RuntimeError(
                "cannot handle multidimensional fweights")
        if fweights.shape[0] != X.shape[1]:
            raise RuntimeError(
                "incompatible numbers of samples and fweights")
        if any(fweights < 0):
            raise ValueError(
                "fweights cannot be negative")
        w = fweights
    if aweights is not None:
        aweights = astensor(aweights, dtype=np.float)
        if aweights.ndim > 1:
            raise RuntimeError(
                "cannot handle multidimensional aweights")
        if aweights.shape[0] != X.shape[1]:
            raise RuntimeError(
                "incompatible numbers of samples and aweights")
        if any(aweights < 0):
            raise ValueError(
                "aweights cannot be negative")
        if w is None:
            w = aweights
        else:
            w *= aweights

    avg, w_sum = average(X, axis=1, weights=w, returned=True)
    w_sum = w_sum[0]

    # Determine the normalization
    if w is None:
        fact = X.shape[1] - ddof
    elif ddof == 0:
        fact = w_sum
    elif aweights is None:
        fact = w_sum - ddof
    else:
        fact = w_sum - ddof * sum(w * aweights) / w_sum

    X -= avg[:, None]
    if w is None:
        X_T = X.T
    else:
        X_T = (X * w).T
    c = dot(X, X_T.conj())
    if isinstance(fact, Tensor):
        fact = where(fact <= 0, 0.0, fact)
        fact = fact.astype(float)
    else:
        if fact <= 0:
            warnings.warn("Degrees of freedom <= 0 for slice",
                          RuntimeWarning, stacklevel=2)
            fact = 0.0
        fact = np.float64(fact)
    c = c * (1. / fact)
    return squeeze(c)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .server import MarsRequestHandler, register_web_handler, get_jinja_env

_jinja_env = get_jinja_env()


class SchedulerListHandler(MarsRequestHandler):
    def get(self):
        schedulers_info = self.web_api.get_schedulers_info()

        template = _jinja_env.get_template('scheduler_pages/list.html')
        self.write(template.render(scheduler_metrics=schedulers_info))


class SchedulerHandler(MarsRequestHandler):
    def get(self, endpoint):
        schedulers_info = self.web_api.get_schedulers_info()

        template = _jinja_env.get_template('scheduler_pages/detail.html')
        self.write(template.render(
            endpoint=endpoint,
            scheduler_metrics=schedulers_info[endpoint],
        ))


register_web_handler('/scheduler', SchedulerListHandler)
register_web_handler('/scheduler/(?P<endpoint>[^/]+)', SchedulerHandler)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
import itertools
from collections import namedtuple

try:
    import lightgbm
except ImportError:
    lightgbm = None


class LGBMModelType(enum.Enum):
    CLASSIFIER = 0
    REGRESSOR = 1
    RANKER = 2


_model_type_to_model = dict()
if lightgbm:
    _model_type_to_model = {
        LGBMModelType.CLASSIFIER: lightgbm.LGBMClassifier,
        LGBMModelType.REGRESSOR: lightgbm.LGBMRegressor,
        LGBMModelType.RANKER: lightgbm.LGBMRanker,
    }


def get_model_cls_from_type(model_type: LGBMModelType):
    return _model_type_to_model[model_type]


TrainTuple = namedtuple('TrainTuple', 'data label sample_weight init_score')


class LGBMScikitLearnBase:
    @staticmethod
    def _copy_extra_params(source, dest):
        params = source.get_params()
        attributes = source.__dict__
        extra_param_names = set(attributes.keys()).difference(params.keys())
        for name in extra_param_names:
            setattr(dest, name, attributes[name])

    @staticmethod
    def _wrap_train_tuple(data, label, sample_weight=None, init_score=None):
        return TrainTuple(data, label, sample_weight, init_score)

    @staticmethod
    def _wrap_eval_tuples(eval_set=None, eval_sample_weight=None, eval_init_score=None):
        if not eval_set:
            return None

        tps = []
        for (data, label), weight, score in zip(eval_set, eval_sample_weight or itertools.repeat(None),
                                                eval_init_score or itertools.repeat(None)):
            tps.append(TrainTuple(data, label, weight, score))
        return tps

    def fit(self, X, y, sample_weight=None, **kwargs):
        raise NotImplementedError

    def predict(self, X, **kwargs):
        raise NotImplementedError

    def predict_proba(self, X, **kwargs):
        raise NotImplementedError



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...config import options
from ...core import OutputType
from ...context import get_context, RunningMode
from ...serialize import TupleField, ListField, StringField
from ...tiles import TilesError
from ...utils import check_chunks_unknown_shape
from ..operands import DataFrameOperand, DataFrameOperandMixin

try:
    import vineyard
except ImportError:
    vineyard = None


class DataFrameToVineyardChunk(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.DATAFRAME_STORE_VINEYARD_CHUNK

    # vineyard ipc socket
    _vineyard_socket = StringField('vineyard_socket')

    def __init__(self, vineyard_socket=None, dtypes=None, **kw):
        super().__init__(_vineyard_socket=vineyard_socket, _dtypes=dtypes, _output_types=[OutputType.dataframe], **kw)

    def __call__(self, df):
        return self.new_dataframe([df], shape=(0, 0), dtypes=df.dtypes,
                                  index_value=df.index_value, columns_value=df.columns_value)

    @property
    def vineyard_socket(self):
        return self._vineyard_socket

    @classmethod
    def _get_out_chunk(cls, op, in_chunk):
        chunk_op = op.copy().reset_key()
        out_chunk_shape = (np.nan,) * in_chunk.ndim
        return chunk_op.new_chunk([in_chunk], shape=out_chunk_shape,
                                  index=in_chunk.index)

    @classmethod
    def tile(cls, op):
        # Not truely necessary, but putting a barrier here will make things simple
        check_chunks_unknown_shape(op.inputs, TilesError)

        in_df = op.inputs[0]

        out_chunks = []
        for chunk in in_df.chunks:
            chunk_op = op.copy().reset_key()
            chunk = chunk_op.new_chunk([chunk], shape=chunk.shape, dtypes=chunk.dtypes,
                                      index_value=chunk.index_value,
                                      columns_value=chunk.columns_value,
                                      index=chunk.index)
            out_chunks.append(chunk)

        new_op = op.copy().reset_key()
        return new_op.new_dataframes(op.inputs, shape=op.inputs[0].shape,
                                     dtypes=in_df.dtypes,
                                     index_value=in_df.index_value,
                                     columns_value=in_df.columns_value,
                                     chunks=out_chunks, nsplits=((np.nan,),))

    @classmethod
    def execute(cls, ctx, op):
        if vineyard is None:
            raise RuntimeError('vineyard is not available')
        client = vineyard.connect(op.vineyard_socket)

        df = vineyard.DataFrameBuilder(client)
        for name, value in ctx[op.inputs[0].key].iteritems():
            np_value = value.to_numpy(copy=False)
            if not np_value.flags['C_CONTIGUOUS']:
                np_value = np.ascontiguousarray(np_value)
            df.add(str(name), vineyard.TensorBuilder.from_numpy(client, np_value))
        df.chunk_index = op.inputs[0].index
        df = df.build(client)
        df.persist(client)

        # store the result object id to execution context
        ctx[op.outputs[0].key] = pd.DataFrame([(client.instance_id, df.id)])


class DataFrameToVineyardChunkMap(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.DATAFRAME_STORE_VINEYARD_CHUNK_MAP

    _local_chunks = ListField("local_chunks")

    # vineyard ipc socket
    _vineyard_socket = StringField('vineyard_socket')

    def __init__(self, vineyard_socket=None, **kw):
        super().__init__(_vineyard_socket=vineyard_socket, _output_types=[OutputType.dataframe], **kw)

    @property
    def local_chunks(self):
        return self._local_chunks

    @property
    def vineyard_socket(self):
        return self._vineyard_socket

    def __call__(self, df):
        return self.new_dataframe([df], shape=df.shape, dtypes=df.dtypes,
                                  index_value=df.index_value, columns_value=df.columns_value)

    @classmethod
    def _process_out_chunks(cls, op, out_chunks):
        merge_op = DataFrameToVinyardStoreGlobalMeta(
                vineyard_socket=op.vineyard_socket,
                chunk_shape=op.inputs[0].chunk_shape,
                shape=op.inputs[0].shape,
                dtypes=op.inputs[0].dtypes)
        return merge_op.new_chunks(out_chunks, shape=out_chunks[0].shape,
                                   index=(0,) * out_chunks[0].ndim)

    @classmethod
    def tile(cls, op):
        check_chunks_unknown_shape(op.inputs, TilesError)

        if vineyard is None:
            raise RuntimeError('vineyard is not available')
        client = vineyard.connect(op.vineyard_socket)

        ctx = get_context()
        if ctx.running_mode == RunningMode.distributed:
            metas = ctx.get_worker_metas()
            workers = {meta['vineyard']['instance_id']: addr for addr, meta in metas.items()}
        else:
            workers = {client.instance_id: '127.0.0.1'}

        all_chunk_ids = ctx.get_chunk_results([c.key for c in op.inputs[0].chunks])
        all_chunk_ids = pd.concat(all_chunk_ids, axis='index')
        chunk_map = defaultdict(list)
        for instance_id, chunk_id in all_chunk_ids.itertuples(index=False):
            chunk_map[instance_id].append(chunk_id)

        out_chunks = []
        for idx, (instance_id, local_chunks) in enumerate(chunk_map.items()):
            chunk_op = op.copy().reset_key()
            chunk_op._local_chunks = local_chunks
            chunk_op._expect_worker = workers[instance_id]
            out_chunks.append(chunk_op.new_chunk(op.inputs[0].chunks, shape=(1,), index=(idx,)))
        out_chunks = cls._process_out_chunks(op, out_chunks)

        new_op = op.copy()
        return new_op.new_dataframes(op.inputs, shape=(len(out_chunks),),
                                     dtypes=op.inputs[0].dtypes,
                                     index_value=op.inputs[0].index_value,
                                     columns_value=op.inputs[0].columns_value,
                                     chunks=out_chunks,
                                     nsplits=((1,) * len(out_chunks),))

    @classmethod
    def execute(cls, ctx, op):
        if vineyard is None:
            raise RuntimeError('vineyard is not available')
        client = vineyard.connect(op.vineyard_socket)
        chunk_blob = vineyard.UInt64VectorBuilder(client, op.local_chunks).build(client)
        ctx[op.outputs[0].key] = (client.instance_id, chunk_blob.id)


class DataFrameToVinyardStoreGlobalMeta(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.DATAFRAME_STORE_VINEYARD_GLOBAL_META

    _shape = TupleField('shape')
    _chunk_shape = TupleField('chunk_shape')

    # vineyard ipc socket
    _vineyard_socket = StringField('vineyard_socket')

    def __init__(self, vineyard_socket=None, chunk_shape=None, shape=None, **kw):
        super().__init__(_vineyard_socket=vineyard_socket,
                         _chunk_shape=chunk_shape, _shape=shape,
                         _output_types=[OutputType.dataframe], **kw)

    @property
    def shape(self):
        return self._shape

    @property
    def chunk_shape(self):
        return self._chunk_shape

    @property
    def vineyard_socket(self):
        return self._vineyard_socket

    @classmethod
    def _process_out_chunks(cls, op, out_chunks):
        if len(out_chunks) == 1:
            return out_chunks
        else:
            raise NotImplementedError('not implemented')

    @classmethod
    def tile(cls, op):
        return [super().tile(op)[0]]

    @classmethod
    def execute(cls, ctx, op):
        if vineyard is None:
            raise RuntimeError('vineyard is not available')
        client = vineyard.connect(op.vineyard_socket)

        df = vineyard.GlobalDataFrameBuilder(client)
        for in_chunk in op.inputs:
            instance_id, chunk_blob_id = ctx[in_chunk.key]
            df.add_chunks(instance_id, chunk_blob_id)
        df.chunk_shape = op.chunk_shape
        df = df.build(client)
        df.persist(client)

        # store the result object id to execution context
        ctx[op.outputs[0].key] = vineyard.ObjectID.stringify(df.id)


def to_vineyard(df, vineyard_socket=None):
    if vineyard_socket is None:
        vineyard_socket = options.vineyard.socket

    chunk_op = DataFrameToVineyardChunk(vineyard_socket=vineyard_socket)
    chunk_map_op = DataFrameToVineyardChunkMap(vineyard_socket=vineyard_socket)
    return chunk_map_op(chunk_op(df))



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from collections.abc import MutableMapping

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...core import OutputType
from ...serialize import KeyField, AnyField, StringField
from ...tiles import TilesError
from ...utils import check_chunks_unknown_shape
from ..core import SERIES_TYPE
from ..operands import DataFrameOperand, DataFrameOperandMixin


class DataFrameMap(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.MAP

    _input = KeyField('input')
    _arg = AnyField('arg')
    _na_action = StringField('na_action')

    def __init__(self, arg=None, na_action=None, output_types=None, **kw):
        super().__init__(_arg=arg, _na_action=na_action,
                         _output_types=output_types, **kw)
        if not self.output_types:
            self.output_types = [OutputType.series]

    @property
    def input(self):
        return self._input

    @property
    def arg(self):
        return self._arg

    @property
    def na_action(self):
        return self._na_action

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]
        if len(inputs) == 2:
            self._arg = self._inputs[1]

    def __call__(self, series, dtype):
        if dtype is None:
            inferred_dtype = None
            if callable(self._arg):
                # arg is a function, try to inspect the signature
                sig = inspect.signature(self._arg)
                return_type = sig.return_annotation
                if return_type is not inspect._empty:
                    inferred_dtype = np.dtype(return_type)
            else:
                if isinstance(self._arg, MutableMapping):
                    inferred_dtype = pd.Series(self._arg).dtype
                else:
                    inferred_dtype = self._arg.dtype
            if inferred_dtype is not None and np.issubdtype(inferred_dtype, np.number):
                if np.issubdtype(inferred_dtype, np.inexact):
                    # for the inexact e.g. float
                    # we can make the decision,
                    # but for int, due to the nan which may occur,
                    # we cannot infer the dtype
                    dtype = inferred_dtype
            else:
                dtype = inferred_dtype

        if dtype is None:
            raise ValueError('cannot infer dtype, '
                             'it needs to be specified manually for `map`')
        else:
            dtype = np.int64 if dtype is int else dtype
            dtype = np.dtype(dtype)

        inputs = [series]
        if isinstance(self._arg, SERIES_TYPE):
            inputs.append(self._arg)
        return self.new_series(inputs, shape=series.shape, dtype=dtype,
                               index_value=series.index_value, name=series.name)

    @classmethod
    def tile(cls, op):
        in_series = op.input
        out_series = op.outputs[0]

        arg = op.arg
        if len(op.inputs) == 2:
            # make sure arg has known shape when it's a md.Series
            check_chunks_unknown_shape([op.arg], TilesError)
            arg = op.arg.rechunk(op.arg.shape)._inplace_tile()

        out_chunks = []
        for chunk in in_series.chunks:
            chunk_op = op.copy().reset_key()
            chunk_inputs = [chunk]
            if len(op.inputs) == 2:
                chunk_inputs.append(arg.chunks[0])
            out_chunk = chunk_op.new_chunk(chunk_inputs, shape=chunk.shape,
                                           dtype=out_series.dtype,
                                           index_value=chunk.index_value,
                                           name=out_series.name,
                                           index=chunk.index)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        params = out_series.params
        params['chunks'] = out_chunks
        params['nsplits'] = in_series.nsplits
        return new_op.new_seriess(op.inputs, kws=[params])

    @classmethod
    def execute(cls, ctx, op):
        series = ctx[op.inputs[0].key]
        out = op.outputs[0]
        if len(op.inputs) == 2:
            arg = ctx[op.inputs[1].key]
        else:
            arg = op.arg

        ret = series.map(arg, na_action=op.na_action)
        if ret.dtype != out.dtype:
            ret = ret.astype(out.dtype)
        ctx[out.key] = ret


def map_(series, arg, na_action=None, dtype=None):
    op = DataFrameMap(arg=arg, na_action=na_action)
    return op(series, dtype=dtype)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numbers import Integral

import numpy as np

from ... import opcodes as OperandDef
from ...config import options
from ...serialize import ValueType, AnyField, KeyField, BoolField, TupleField
from ...tiles import TilesError
from ...utils import check_chunks_unknown_shape, ceildiv, recursive_tile
from ..operands import TensorOperandMixin
from ..core import TENSOR_TYPE, TENSOR_CHUNK_TYPE, TensorOrder
from ..datasource import arange, array
from ..utils import decide_chunk_sizes, normalize_chunk_sizes, gen_random_seeds
from ..array_utils import as_same_device, device
from .core import TensorRandomOperand, RandomState


class TensorChoice(TensorRandomOperand, TensorOperandMixin):
    _op_type_ = OperandDef.RAND_CHOICE

    _a = AnyField('a')
    _size = TupleField('size', ValueType.int64)
    _replace = BoolField('replace')
    _p = KeyField('p')

    def __init__(self, a=None, size=None, replace=None, p=None,
                 state=None, seed=None, dtype=None, gpu=None, **kw):
        super().__init__(_a=a, _size=size, _replace=replace, _p=p,
                         _state=state, _seed=seed, _dtype=dtype, _gpu=gpu, **kw)

    @property
    def a(self):
        return self._a

    @property
    def size(self):
        return self._size

    @property
    def replace(self):
        return self._replace

    @property
    def p(self):
        return self._p

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if isinstance(self._a, (TENSOR_TYPE, TENSOR_CHUNK_TYPE)):
            self._a = self._inputs[0]
        if isinstance(self._p, (TENSOR_TYPE, TENSOR_CHUNK_TYPE)):
            self._p = self._inputs[-1]

    def __call__(self, a, p, chunk_size=None):
        inputs = []
        if isinstance(a, TENSOR_TYPE):
            inputs.append(a)
        if isinstance(p, TENSOR_TYPE):
            inputs.append(p)
        return self.new_tensor(inputs, shape=self._size,
                               raw_chunk_size=chunk_size,
                               order=TensorOrder.C_ORDER)

    @classmethod
    def _tile_one_chunk(cls, op, a, p):
        out = op.outputs[0]
        chunk_op = op.copy().reset_key()
        chunk_op._seed = gen_random_seeds(1, op.state)[0]
        chunk_inputs = []
        if isinstance(a, TENSOR_TYPE):
            chunk_op._a = a.chunks[0]
            chunk_inputs.append(chunk_op.a)
        else:
            chunk_op._a = a
        if isinstance(p, TENSOR_TYPE):
            chunk_op._p = p.chunks[0]
            chunk_inputs.append(chunk_op.p)
        else:
            chunk_op._p = p
        chunk = chunk_op.new_chunk(chunk_inputs, shape=out.shape,
                                   index=(0,) * out.ndim,
                                   order=out.order)

        new_op = op.copy()
        return new_op.new_tensors(op.inputs, shape=out.shape,
                                  order=out.order, chunks=[chunk],
                                  nsplits=tuple((s,) for s in out.shape))

    @classmethod
    def _tile_sample_with_replacement(cls, op, a, nsplits):
        out_shape = tuple(sum(ns) for ns in nsplits)
        out_size = np.prod(out_shape).item()
        most_chunk_size = np.prod([max(ns) for ns in nsplits]).item()

        is_a_int = False
        if isinstance(a, Integral):
            is_a_int = True
            a_size = a
        else:
            a = array(a)
            a_size = a.size

        rs = RandomState.from_numpy(op.state)

        if is_a_int:
            # the indices is just the result
            ret = rs.randint(a_size, size=out_shape, chunk_size=nsplits)
        else:
            # gen indices first, need to be flattened
            indices = rs.randint(a_size, size=out_size, chunk_size=most_chunk_size)
            # get result via fancy indexing
            ret = a[indices]
            if len(out_shape) > 1:
                # reshape back if out's ndim > 1
                ret = ret.reshape(out_shape)
            ret = ret.rechunk(nsplits)

        return [recursive_tile(ret)]

    @classmethod
    def _tile_sample_without_replacement(cls, op, a, nsplits):
        from ..base import searchsorted
        from ..merge.stack import TensorStack
        from ..indexing.getitem import TensorIndex

        out = op.outputs[0]
        out_shape = tuple(sum(ns) for ns in nsplits)
        # to sample count
        m = np.prod(out_shape).item()

        if isinstance(a, Integral):
            a_size = a
            a = arange(a)
        else:
            a = array(a)
            a_size = a.size
        a = a._inplace_tile()

        if any(cs < m for cs in a.nsplits[0]):
            # make sure all chunk > m
            n_chunk = min(max(a.size // (m + 1), 1), a.chunk_shape[0])
            chunk_size = ceildiv(a.size, n_chunk)
            chunk_sizes = normalize_chunk_sizes(a.size, chunk_size)[0]
            if chunk_sizes[-1] < m and len(chunk_sizes) > 1:
                # the last chunk may still less than m
                # merge it into previous one
                chunk_sizes[-2] += chunk_sizes[-1]
                chunk_sizes = chunk_sizes[:-1]
            a = a.rechunk({0: chunk_sizes})._inplace_tile()
            if len(chunk_sizes) == 1:
                return cls._tile_one_chunk(op, a, None)

        # for each chunk in a, do regular sampling
        sampled_chunks = []
        sample_seeds = gen_random_seeds(len(a.chunks), op.state)
        for seed, chunk in zip(sample_seeds, a.chunks):
            chunk_op = op.copy().reset_key()
            chunk_op._a = chunk
            chunk_op._size = (m,)
            chunk_op._seed = seed
            sampled_chunk = chunk_op.new_chunk([chunk], shape=(m,),
                                               order=out.order,
                                               index=chunk.index)
            sampled_chunks.append(sampled_chunk)

        if len(sampled_chunks) == 1:
            out_chunk = sampled_chunks[0]
        else:
            stacked_chunk = TensorStack(axis=0, dtype=sampled_chunks[0].dtype).new_chunk(
                sampled_chunks, shape=(len(a.chunks), m), order=TensorOrder.C_ORDER)

            # gen indices with length m from 0...a.size
            state = RandomState.from_numpy(op.state)
            indices = state.randint(a_size, size=(m,))
            cum_offsets = np.cumsum(a.nsplits[0])
            ind = recursive_tile(searchsorted(cum_offsets, indices, side='right'))
            ind_chunk = ind.chunks[0]

            # do fancy index to find result
            indexes = (ind_chunk, arange(m)._inplace_tile().chunks[0])
            out_chunk = TensorIndex(dtype=stacked_chunk.dtype, indexes=indexes).new_chunk(
                [stacked_chunk] + list(indexes), shape=(m,), order=TensorOrder.C_ORDER)

        ret = op.copy().new_tensor(op.inputs, shape=(m,), order=out.order,
                                   nsplits=((m,),), chunks=[out_chunk])
        if len(out_shape) > 0:
            ret = ret.reshape(out_shape)._inplace_tile()
        ret = ret.rechunk(nsplits)._inplace_tile()
        return [recursive_tile(ret)]

    @classmethod
    def tile(cls, op):
        check_chunks_unknown_shape(op.inputs, TilesError)

        out = op.outputs[0]
        chunk_size = out.extra_params.raw_chunk_size or options.chunk_size
        nsplits = decide_chunk_sizes(out.shape, chunk_size, out.dtype.itemsize)
        inputs = op.inputs

        a, p = op.a, op.p
        if p is not None:
            # we cannot handle p in a parallel fashion
            inputs = []
            if isinstance(a, TENSOR_TYPE):
                a = a.rechunk(a.shape)._inplace_tile()
                inputs.append(a)
            p = p.rechunk(p.shape)._inplace_tile()
            inputs.append(p)

            # ignore nsplits if p is specified
            nsplits = ((s,) for s in out.shape)

        # all inputs and outputs has 1 chunk
        if all(len(inp.chunks) == 1 for inp in inputs) and \
                all(len(ns) == 1 for ns in nsplits):
            return cls._tile_one_chunk(op, a, p)

        if op.replace:
            return cls._tile_sample_with_replacement(op, a, nsplits)
        else:
            return cls._tile_sample_without_replacement(op, a, nsplits)

    @classmethod
    def execute(cls, ctx, op):
        inputs, device_id, xp = as_same_device(
            [ctx[inp.key] for inp in op.inputs], device=op.device, ret_extra=True)
        if isinstance(op.a, TENSOR_CHUNK_TYPE):
            a = inputs[0]
        else:
            a = op.a
        if isinstance(op.p, TENSOR_CHUNK_TYPE):
            p = inputs[-1]
        else:
            p = op.p

        with device(device_id):
            rs = xp.random.RandomState(op.seed)
            ctx[op.outputs[0].key] = rs.choice(
                a, size=op.size, replace=op.replace, p=p)


def choice(random_state, a, size=None, replace=True, p=None, chunk_size=None, gpu=None):
    """
    Generates a random sample from a given 1-D array

    Parameters
    -----------
    a : 1-D array-like or int
        If a tensor, a random sample is generated from its elements.
        If an int, the random sample is generated as if a were mt.arange(a)
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  Default is None, in which case a
        single value is returned.
    replace : boolean, optional
        Whether the sample is with or without replacement
    p : 1-D array-like, optional
        The probabilities associated with each entry in a.
        If not given the sample assumes a uniform distribution over all
        entries in a.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    --------
    samples : single item or tensor
        The generated random samples

    Raises
    -------
    ValueError
        If a is an int and less than zero, if a or p are not 1-dimensional,
        if a is an array-like of size 0, if p is not a vector of
        probabilities, if a and p have different lengths, or if
        replace=False and the sample size is greater than the population
        size

    See Also
    ---------
    randint, shuffle, permutation

    Examples
    ---------
    Generate a uniform random sample from mt.arange(5) of size 3:

    >>> import mars.tensor as mt

    >>> mt.random.choice(5, 3).execute()
    array([0, 3, 4])
    >>> #This is equivalent to mt.random.randint(0,5,3)

    Generate a non-uniform random sample from np.arange(5) of size 3:

    >>> mt.random.choice(5, 3, p=[0.1, 0, 0.3, 0.6, 0]).execute()
    array([3, 3, 0])

    Generate a uniform random sample from mt.arange(5) of size 3 without
    replacement:

    >>> mt.random.choice(5, 3, replace=False).execute()
    array([3,1,0])
    >>> #This is equivalent to np.random.permutation(np.arange(5))[:3]

    Generate a non-uniform random sample from mt.arange(5) of size
    3 without replacement:

    >>> mt.random.choice(5, 3, replace=False, p=[0.1, 0, 0.3, 0.6, 0]).execute()
    array([2, 3, 0])

    Any of the above can be repeated with an arbitrary array-like
    instead of just integers. For instance:

    >>> aa_milne_arr = ['pooh', 'rabbit', 'piglet', 'Christopher']
    >>> np.random.choice(aa_milne_arr, 5, p=[0.5, 0.1, 0.1, 0.3])
    array(['pooh', 'pooh', 'pooh', 'Christopher', 'piglet'],
          dtype='|S11')
    """

    if isinstance(a, Integral):
        if a <= 0:
            raise ValueError('a must be greater than 0')
        a_size = a
        dtype = np.random.choice(1, size=(), p=np.array([1]) if p is not None else p).dtype
    else:
        a = array(a)
        if a.ndim != 1:
            raise ValueError('a must be one dimensional')
        a_size = a.size
        dtype = a.dtype

    if p is not None:
        if not isinstance(p, TENSOR_TYPE):
            p = np.asarray(p)
            if not np.isclose(p.sum(), 1, rtol=1e-7, atol=0):
                raise ValueError('probabilities do not sum to 1')
            p = array(p, chunk_size=p.size)
        if p.ndim != 1:
            raise ValueError('p must be one dimensional')

    if size is None:
        size = ()
        length = 1
    else:
        try:
            tuple(size)
            length = np.prod(size)
        except TypeError:
            length = size
    if replace is False and length > a_size:
        raise ValueError("Cannot take a larger sample than population when 'replace=False'")

    size = random_state._handle_size(size)
    op = TensorChoice(a=a, p=p, state=random_state.to_numpy(),
                      replace=replace, size=size, dtype=dtype, gpu=gpu)
    return op(a, p, chunk_size=chunk_size)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mars/serialize/protos/operand.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from mars.serialize.protos import value_pb2 as mars_dot_serialize_dot_protos_dot_value__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='mars/serialize/protos/operand.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n#mars/serialize/protos/operand.proto\x1a!mars/serialize/protos/value.proto\"\xaf\x38\n\nOperandDef\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\n\n\x02id\x18\x02 \x01(\t\x12#\n\x04\x61ttr\x18\x03 \x03(\x0b\x32\x15.OperandDef.AttrEntry\x1a\x33\n\tAttrEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x15\n\x05value\x18\x02 \x01(\x0b\x32\x06.Value:\x02\x38\x01\"\xac\x37\n\x0bOperandType\x12\x08\n\x04NULL\x10\x00\x12\n\n\x06SCALAR\x10\x01\x12\x16\n\x12TENSOR_DATA_SOURCE\x10\x02\x12\x0f\n\x0bTENSOR_ONES\x10\x03\x12\x14\n\x10TENSOR_ONES_LIKE\x10\x04\x12\x10\n\x0cTENSOR_ZEROS\x10\x05\x12\x15\n\x11TENSOR_ZEROS_LIKE\x10\x06\x12\x10\n\x0cTENSOR_EMPTY\x10\x07\x12\x15\n\x11TENSOR_EMPTY_LIKE\x10\x08\x12\x0f\n\x0bTENSOR_FULL\x10\t\x12\x14\n\x10TENSOR_FULL_LIKE\x10\x19\x12\x11\n\rTENSOR_ARANGE\x10\n\x12\x12\n\x0eTENSOR_INDICES\x10\x0b\x12\x0f\n\x0bTENSOR_DIAG\x10\x0c\x12\x0e\n\nTENSOR_EYE\x10\r\x12\x13\n\x0fTENSOR_LINSPACE\x10\x0e\x12\x0f\n\x0bTENSOR_TRIU\x10\x0f\x12\x0f\n\x0bTENSOR_TRIL\x10\x10\x12\x16\n\x12TENSOR_FROM_TILEDB\x10\x12\x12\x17\n\x13TENSOR_STORE_TILEDB\x10\x13\x12#\n\x1fTENSOR_STORE_TILEDB_CONSOLIDATE\x10\x14\x12\x19\n\x15TENSOR_FROM_DATAFRAME\x10\x16\x12\x14\n\x10TENSOR_FROM_HDF5\x10\x1b\x12\x15\n\x11TENSOR_STORE_HDF5\x10\x1c\x12\x14\n\x10TENSOR_FROM_ZARR\x10\x1d\x12\x15\n\x11TENSOR_STORE_ZARR\x10 \x12\x19\n\x15\x44\x41TAFRAME_DATA_SOURCE\x10\x11\x12\x19\n\x15\x44\x41TAFRAME_FROM_TENSOR\x10\x15\x12\x16\n\x12SERIES_FROM_TENSOR\x10\x1a\x12\x1a\n\x16\x44\x41TAFRAME_FROM_RECORDS\x10\x18\x12\x16\n\x12SERIES_DATA_SOURCE\x10\x17\x12\x15\n\x11INDEX_DATA_SOURCE\x10!\x12\x0e\n\nDATE_RANGE\x10\"\x12\x13\n\x0fTIMEDELTA_RANGE\x10#\x12\x10\n\x0cMEMORY_USAGE\x10$\x12\n\n\x06TO_GPU\x10\x1e\x12\n\n\x06TO_CPU\x10\x1f\x12\r\n\tRAND_RAND\x10)\x12\x0e\n\nRAND_RANDN\x10*\x12\x10\n\x0cRAND_RANDINT\x10+\x12\x18\n\x14RAND_RANDOM_INTEGERS\x10,\x12\x16\n\x12RAND_RANDOM_SAMPLE\x10-\x12\x0f\n\x0bRAND_RANDOM\x10.\x12\r\n\tRAND_RANF\x10/\x12\x0f\n\x0bRAND_SAMPLE\x10\x30\x12\x0e\n\nRAND_BYTES\x10\x31\x12\r\n\tRAND_BETA\x10\x32\x12\x11\n\rRAND_BINOMIAL\x10\x33\x12\x12\n\x0eRAND_CHISQUARE\x10\x34\x12\x0f\n\x0bRAND_CHOICE\x10\x35\x12\x12\n\x0eRAND_DIRICHLET\x10\x36\x12\x14\n\x10RAND_EXPONENTIAL\x10\x37\x12\n\n\x06RAND_F\x10\x38\x12\x0e\n\nRAND_GAMMA\x10\x39\x12\x12\n\x0eRAND_GEOMETRIC\x10:\x12\x0f\n\x0bRAND_GUMBEL\x10;\x12\x17\n\x13RAND_HYPERGEOMETRIC\x10<\x12\x10\n\x0cRAND_LAPLACE\x10=\x12\x11\n\rRAND_LOGISTIC\x10>\x12\x12\n\x0eRAND_LOGNORMAL\x10?\x12\x12\n\x0eRAND_LOGSERIES\x10@\x12\x14\n\x10RAND_MULTINOMIAL\x10\x41\x12\x1c\n\x18RAND_MULTIVARIATE_NORMAL\x10\x42\x12\x1a\n\x16RAND_NEGATIVE_BINOMIAL\x10\x43\x12\x1c\n\x18RAND_NONCENTRAL_CHISQURE\x10\x44\x12\x15\n\x11RAND_NONCENTRAL_F\x10\x45\x12\x0f\n\x0bRAND_NORMAL\x10\x46\x12\x0f\n\x0bRAND_PARETO\x10G\x12\x14\n\x10RAND_PERMUTATION\x10H\x12\x10\n\x0cRAND_POSSION\x10I\x12\x0e\n\nRAND_POWER\x10J\x12\x11\n\rRAND_RAYLEIGH\x10K\x12\x10\n\x0cRAND_SHUFFLE\x10L\x12\x18\n\x14RAND_STANDARD_CAUCHY\x10M\x12\x1d\n\x19RAND_STANDARD_EXPONENTIAL\x10N\x12\x18\n\x14RAND_STANDARD_GAMMMA\x10O\x12\x18\n\x14RAND_STANDARD_NORMAL\x10P\x12\x13\n\x0fRAND_STANDARD_T\x10Q\x12\x11\n\rRAND_TOMAXINT\x10R\x12\x13\n\x0fRAND_TRIANGULAR\x10S\x12\x10\n\x0cRAND_UNIFORM\x10T\x12\x11\n\rRAND_VONMISES\x10U\x12\r\n\tRAND_WALD\x10V\x12\x10\n\x0cRAND_WEIBULL\x10W\x12\r\n\tRAND_ZIPF\x10X\x12\x0f\n\x0bPERMUTATION\x10Y\x12\n\n\x06UNIQUE\x10Z\x12\x07\n\x03\x41\x44\x44\x10\x65\x12\x07\n\x03SUB\x10\x66\x12\x07\n\x03MUL\x10g\x12\x07\n\x03\x44IV\x10h\x12\x0b\n\x07TRUEDIV\x10i\x12\x0c\n\x08\x46LOORDIV\x10j\x12\x07\n\x03POW\x10k\x12\x07\n\x03MOD\x10l\x12\x08\n\x04\x46MOD\x10m\x12\r\n\tLOGADDEXP\x10n\x12\x0e\n\nLOGADDEXP2\x10o\x12\x0c\n\x08NEGATIVE\x10p\x12\x0c\n\x08POSITIVE\x10q\x12\x0c\n\x08\x41\x42SOLUTE\x10r\x12\x08\n\x04\x46\x41\x42S\x10s\x12\x07\n\x03\x41\x42S\x10t\x12\x08\n\x04RINT\x10u\x12\x08\n\x04SIGN\x10v\x12\x08\n\x04\x43ONJ\x10w\x12\x07\n\x03\x45XP\x10x\x12\x08\n\x04\x45XP2\x10y\x12\x07\n\x03LOG\x10z\x12\x08\n\x04LOG2\x10{\x12\t\n\x05LOG10\x10|\x12\t\n\x05\x45XPM1\x10}\x12\t\n\x05LOG1P\x10~\x12\x08\n\x04SQRT\x10\x7f\x12\x0b\n\x06SQUARE\x10\x80\x01\x12\t\n\x04\x43\x42RT\x10\x81\x01\x12\x0f\n\nRECIPROCAL\x10\x82\x01\x12\x07\n\x02\x45Q\x10\x83\x01\x12\x07\n\x02NE\x10\x84\x01\x12\x07\n\x02LT\x10\x85\x01\x12\x07\n\x02LE\x10\x86\x01\x12\x07\n\x02GT\x10\x87\x01\x12\x07\n\x02GE\x10\x88\x01\x12\x08\n\x03SIN\x10\x89\x01\x12\x08\n\x03\x43OS\x10\x8a\x01\x12\x08\n\x03TAN\x10\x8b\x01\x12\x0b\n\x06\x41RCSIN\x10\x8c\x01\x12\x0b\n\x06\x41RCCOS\x10\x8d\x01\x12\x0b\n\x06\x41RCTAN\x10\x8e\x01\x12\x0c\n\x07\x41RCTAN2\x10\x8f\x01\x12\n\n\x05HYPOT\x10\x90\x01\x12\t\n\x04SINH\x10\x91\x01\x12\t\n\x04\x43OSH\x10\x92\x01\x12\t\n\x04TANH\x10\x93\x01\x12\x0c\n\x07\x41RCSINH\x10\x94\x01\x12\x0c\n\x07\x41RCCOSH\x10\x95\x01\x12\x0c\n\x07\x41RCTANH\x10\x96\x01\x12\x0c\n\x07\x44\x45G2RAD\x10\x97\x01\x12\x0c\n\x07RAD2DEG\x10\x98\x01\x12\x0b\n\x06\x42ITAND\x10\x99\x01\x12\n\n\x05\x42ITOR\x10\x9a\x01\x12\x0b\n\x06\x42ITXOR\x10\x9b\x01\x12\x0b\n\x06INVERT\x10\x9c\x01\x12\x0b\n\x06LSHIFT\x10\x9d\x01\x12\x0b\n\x06RSHIFT\x10\x9e\x01\x12\x08\n\x03\x41ND\x10\x9f\x01\x12\x07\n\x02OR\x10\xa0\x01\x12\x08\n\x03XOR\x10\xa1\x01\x12\x08\n\x03NOT\x10\xa2\x01\x12\x0c\n\x07MAXIMUM\x10\xa3\x01\x12\x0c\n\x07MINIMUM\x10\xa4\x01\x12\x0b\n\x06\x41ROUND\x10\xa5\x01\x12\x10\n\x0b\x46LOAT_POWER\x10\xa6\x01\x12\t\n\x04\x46MAX\x10\xa7\x01\x12\t\n\x04\x46MIN\x10\xa8\x01\x12\r\n\x08ISFINITE\x10\xa9\x01\x12\n\n\x05ISINF\x10\xaa\x01\x12\n\n\x05ISNAN\x10\xab\x01\x12\x0c\n\x07SIGNBIT\x10\xac\x01\x12\r\n\x08\x43OPYSIGN\x10\xad\x01\x12\x0e\n\tNEXTAFTER\x10\xae\x01\x12\x0c\n\x07SPACING\x10\xaf\x01\x12\n\n\x05LDEXP\x10\xb0\x01\x12\n\n\x05\x46REXP\x10\xb1\x01\x12\t\n\x04MODF\x10\xb2\x01\x12\n\n\x05\x46LOOR\x10\xb3\x01\x12\t\n\x04\x43\x45IL\x10\xb4\x01\x12\n\n\x05TRUNC\x10\xb5\x01\x12\x0c\n\x07\x44\x45GREES\x10\xb6\x01\x12\x0c\n\x07RADIANS\x10\xb7\x01\x12\t\n\x04\x43LIP\x10\xb8\x01\x12\x0b\n\x06ISREAL\x10\xb9\x01\x12\x0e\n\tISCOMPLEX\x10\xba\x01\x12\t\n\x04REAL\x10\xbb\x01\x12\t\n\x04IMAG\x10\xbc\x01\x12\x08\n\x03\x46IX\x10\xbd\x01\x12\x07\n\x02I0\x10\xbe\x01\x12\t\n\x04SINC\x10\xbf\x01\x12\x0f\n\nNAN_TO_NUM\x10\xc0\x01\x12\x0c\n\x07ISCLOSE\x10\xc1\x01\x12\x0b\n\x06\x44IVMOD\x10\xc2\x01\x12\n\n\x05\x41NGLE\x10\xc3\x01\x12\r\n\x08SET_REAL\x10\xc4\x01\x12\r\n\x08SET_IMAG\x10\xc5\x01\x12\x0c\n\x07GAMMALN\x10\xc8\x01\x12\x08\n\x03\x45RF\x10\xc9\x01\x12\t\n\x04\x45NTR\x10\xca\x01\x12\r\n\x08REL_ENTR\x10\xcb\x01\x12\n\n\x05PDIST\x10\xe7\x01\x12\n\n\x05\x43\x44IST\x10\xe8\x01\x12\x0f\n\nSQUAREFORM\x10\xe9\x01\x12\r\n\x08TREE_ADD\x10\xfb\x01\x12\x12\n\rTREE_MULTIPLY\x10\xfc\x01\x12\x0b\n\x06\x43UMSUM\x10\xad\x02\x12\x0c\n\x07\x43UMPROD\x10\xae\x02\x12\t\n\x04PROD\x10\xaf\x02\x12\x08\n\x03SUM\x10\xb0\x02\x12\x08\n\x03MAX\x10\xb1\x02\x12\x08\n\x03MIN\x10\xb2\x02\x12\x08\n\x03\x41LL\x10\xb3\x02\x12\x08\n\x03\x41NY\x10\xb4\x02\x12\t\n\x04MEAN\x10\xb5\x02\x12\x0b\n\x06\x41RGMAX\x10\xb6\x02\x12\x0b\n\x06\x41RGMIN\x10\xb7\x02\x12\x0b\n\x06NANSUM\x10\xb8\x02\x12\x0b\n\x06NANMAX\x10\xb9\x02\x12\x0b\n\x06NANMIN\x10\xba\x02\x12\x0c\n\x07NANPROD\x10\xbb\x02\x12\x0c\n\x07NANMEAN\x10\xbc\x02\x12\x0e\n\tNANARGMAX\x10\xbd\x02\x12\x0e\n\tNANARGMIN\x10\xbe\x02\x12\x12\n\rCOUNT_NONZERO\x10\xbf\x02\x12\x0b\n\x06MOMENT\x10\xc0\x02\x12\x0e\n\tNANMOMENT\x10\xc1\x02\x12\x08\n\x03VAR\x10\xc2\x02\x12\x08\n\x03STD\x10\xc3\x02\x12\x0b\n\x06NANVAR\x10\xc4\x02\x12\x0b\n\x06NANSTD\x10\xc5\x02\x12\x0e\n\tNANCUMSUM\x10\xc6\x02\x12\x0f\n\nNANCUMPROD\x10\xc7\x02\x12\n\n\x05\x43OUNT\x10\xd7\x02\x12\x0b\n\x06\x43UMMAX\x10\xd8\x02\x12\x0b\n\x06\x43UMMIN\x10\xd9\x02\x12\r\n\x08\x43UMCOUNT\x10\xda\x02\x12\x0c\n\x07RESHAPE\x10\x91\x03\x12\n\n\x05SLICE\x10\x92\x03\x12\n\n\x05INDEX\x10\x93\x03\x12\x12\n\rINDEXSETVALUE\x10\x94\x03\x12\x10\n\x0b\x43ONCATENATE\x10\x95\x03\x12\x0c\n\x07RECHUNK\x10\x96\x03\x12\x0b\n\x06\x41STYPE\x10\x97\x03\x12\x0e\n\tTRANSPOSE\x10\x98\x03\x12\r\n\x08SWAPAXES\x10\x99\x03\x12\x11\n\x0c\x42ROADCAST_TO\x10\x9a\x03\x12\n\n\x05STACK\x10\x9b\x03\x12\n\n\x05WHERE\x10\x9c\x03\x12\x0b\n\x06\x43HOOSE\x10\x9d\x03\x12\x0c\n\x07NONZERO\x10\x9e\x03\x12\r\n\x08\x41RGWHERE\x10\x9f\x03\x12\x12\n\rUNRAVEL_INDEX\x10\xa0\x03\x12\x16\n\x11RAVEL_MULTI_INDEX\x10\xa1\x03\x12\x10\n\x0b\x41RRAY_SPLIT\x10\xa2\x03\x12\x0c\n\x07SQUEEZE\x10\xa3\x03\x12\r\n\x08\x44IGITIZE\x10\xa4\x03\x12\x0b\n\x06REPEAT\x10\xa5\x03\x12\x0b\n\x06\x43OPYTO\x10\xa6\x03\x12\t\n\x04ISIN\x10\xa7\x03\x12\x11\n\x0cSEARCHSORTED\x10\xac\x03\x12\t\n\x04SORT\x10\xad\x03\x12\x0e\n\tHISTOGRAM\x10\xae\x03\x12\x18\n\x13HISTOGRAM_BIN_EDGES\x10\xaf\x03\x12\x0e\n\tPARTITION\x10\xb0\x03\x12\r\n\x08QUANTILE\x10\xb8\x03\x12\x12\n\rFILL_DIAGONAL\x10\xb9\x03\x12\x0e\n\tNORMALIZE\x10\xba\x03\x12\t\n\x04TOPK\x10\xbb\x03\x12\n\n\x05TRAPZ\x10\xbc\x03\x12\x0e\n\tGET_SHAPE\x10\xbd\x03\x12\x1b\n\x16\x46\x41NCY_INDEX_DISTRIBUTE\x10\xa8\x03\x12\x17\n\x12\x46\x41NCY_INDEX_CONCAT\x10\xa9\x03\x12\x0e\n\tTENSORDOT\x10\xf5\x03\x12\x08\n\x03\x44OT\x10\xf6\x03\x12\x0b\n\x06MATMUL\x10\xf7\x03\x12\r\n\x08\x43HOLESKY\x10\xfe\x03\x12\x07\n\x02QR\x10\xff\x03\x12\x08\n\x03SVD\x10\x80\x04\x12\x07\n\x02LU\x10\x81\x04\x12\x15\n\x10SOLVE_TRIANGULAR\x10\x88\x04\x12\x08\n\x03INV\x10\x89\x04\x12\t\n\x04NORM\x10\x92\x04\x12\x08\n\x03\x46\x46T\x10\xd9\x04\x12\t\n\x04IFFT\x10\xda\x04\x12\t\n\x04\x46\x46T2\x10\xdb\x04\x12\n\n\x05IFFT2\x10\xdc\x04\x12\t\n\x04\x46\x46TN\x10\xdd\x04\x12\n\n\x05IFFTN\x10\xde\x04\x12\t\n\x04RFFT\x10\xdf\x04\x12\n\n\x05IRFFT\x10\xe0\x04\x12\n\n\x05RFFT2\x10\xe1\x04\x12\x0b\n\x06IRFFT2\x10\xe2\x04\x12\n\n\x05RFFTN\x10\xe3\x04\x12\x0b\n\x06IRFFTN\x10\xe4\x04\x12\t\n\x04HFFT\x10\xe5\x04\x12\n\n\x05IHFFT\x10\xe6\x04\x12\x0c\n\x07\x46\x46TFREQ\x10\xe7\x04\x12\x12\n\rFFTFREQ_CHUNK\x10\xe8\x04\x12\r\n\x08RFFTFREQ\x10\xe9\x04\x12\r\n\x08\x46\x46TSHIFT\x10\xea\x04\x12\x0e\n\tIFFTSHIFT\x10\xeb\x04\x12\x0b\n\x06\x45INSUM\x10\xf6\x04\x12\x1e\n\x19SPARSE_MATRIX_DATA_SOURCE\x10\xbd\x05\x12\x14\n\x0f\x44\x45NSE_TO_SPARSE\x10\xbe\x05\x12\x14\n\x0fSPARSE_TO_DENSE\x10\xbf\x05\x12\x08\n\x03MAP\x10\xc6\x05\x12\r\n\x08\x44\x45SCRIBE\x10\xc8\x05\x12\x0c\n\x07\x46ILL_NA\x10\xc9\x05\x12\x0e\n\tAGGREGATE\x10\xca\x05\x12\x12\n\rSTRING_METHOD\x10\xcb\x05\x12\x14\n\x0f\x44\x41TETIME_METHOD\x10\xcc\x05\x12\n\n\x05\x41PPLY\x10\xcd\x05\x12\x0e\n\tTRANSFORM\x10\xce\x05\x12\r\n\x08\x43HECK_NA\x10\xcf\x05\x12\x0c\n\x07\x44ROP_NA\x10\xd0\x05\x12\x0c\n\x07NUNIQUE\x10\xd1\x05\x12\x08\n\x03\x43UT\x10\xd2\x05\x12\n\n\x05SHIFT\x10\xd3\x05\x12\x13\n\x0e\x44\x41TAFRAME_DIFF\x10\xd4\x05\x12\x11\n\x0cVALUE_COUNTS\x10\xd5\x05\x12\x10\n\x0bTO_DATETIME\x10\xd6\x05\x12\x13\n\x0e\x44\x41TAFRAME_DROP\x10\xd7\x05\x12\x14\n\x0f\x44ROP_DUPLICATES\x10\xd8\x05\x12\t\n\x04MELT\x10\xd9\x05\x12\x0e\n\tSET_LABEL\x10\xda\x05\x12\x0b\n\x06RENAME\x10\xdb\x05\x12\x0b\n\x06INSERT\x10\xdc\x05\x12\t\n\x04\x46USE\x10\xa1\x06\x12\n\n\x05\x45NTER\x10\x85\x07\x12\n\n\x05LEAVE\x10\x86\x07\x12\x0f\n\nFIX_LATEST\x10\x87\x07\x12\x0c\n\x07IF_ELSE\x10\x88\x07\x12\x0e\n\tNEXT_ITER\x10\x89\x07\x12\x0e\n\tTABLE_COO\x10\xeb\x07\x12\x0e\n\tSTORE_COO\x10\xec\x07\x12\x12\n\rSHUFFLE_PROXY\x10\xd1\x0f\x12\x1a\n\x15\x44\x41TAFRAME_INDEX_ALIGN\x10\xd4\x0f\x12\x1a\n\x15UNIQUE_INVERSE_REDUCE\x10\xdd\x0f\x12\x18\n\x13\x44\x41TAFRAME_SET_INDEX\x10\xe4\x0f\x12\x1b\n\x16\x44\x41TAFRAME_ILOC_GETITEM\x10\xe5\x0f\x12\x1b\n\x16\x44\x41TAFRAME_ILOC_SETITEM\x10\xe6\x0f\x12\x1a\n\x15\x44\x41TAFRAME_LOC_GETITEM\x10\xe7\x0f\x12\x1a\n\x15\x44\x41TAFRAME_LOC_SETITEM\x10\xe8\x0f\x12\x1c\n\x17\x44\x41TAFRAME_SHUFFLE_MERGE\x10\xda\x0f\x12\"\n\x1d\x44\x41TAFRAME_SHUFFLE_MERGE_ALIGN\x10\xdb\x0f\x12\x0b\n\x06\x41PPEND\x10\xdf\x0f\x12\x10\n\x0bRESET_INDEX\x10\xec\x0f\x12\x0c\n\x07REINDEX\x10\xed\x0f\x12\x0c\n\x07GROUPBY\x10\xee\x0f\x12\x10\n\x0bGROUPBY_AGG\x10\xf1\x0f\x12\x13\n\x0eGROUPBY_CONCAT\x10\xf2\x0f\x12\x1d\n\x18PSRS_SORT_REGULAR_SMAPLE\x10\xf8\x0f\x12\x16\n\x11PSRS_CONCAT_PIVOT\x10\xf9\x0f\x12\x11\n\x0cPSRS_SHUFFLE\x10\xfa\x0f\x12\x0f\n\nPSRS_ALIGN\x10\xfb\x0f\x12\x19\n\x14\x43\x41LC_PARTITIONS_INFO\x10\xfe\x0f\x12\x15\n\x10PARTITION_MERGED\x10\xff\x0f\x12\x10\n\x0bSORT_VALUES\x10\x82\x10\x12\x0f\n\nSORT_INDEX\x10\x83\x10\x12\x10\n\x0bROLLING_AGG\x10\x8c\x10\x12\x12\n\rEXPANDING_AGG\x10\x8d\x10\x12\x0c\n\x07\x45WM_AGG\x10\x8e\x10\x12\r\n\x08READ_CSV\x10\xb4\x10\x12\x0b\n\x06TO_CSV\x10\xb5\x10\x12\x11\n\x0cREAD_PARQUET\x10\xb7\x10\x12\x0f\n\nTO_PARQUET\x10\xb8\x10\x12\r\n\x08READ_SQL\x10\xb9\x10\x12\x0b\n\x06TO_SQL\x10\xbc\x10\x12\x10\n\x0bTO_CSV_STAT\x10\xb6\x10\x12\x1c\n\x17STANDARDIZE_RANGE_INDEX\x10\xbb\x10\x12\x19\n\x14SUCCESSORS_EXCLUSIVE\x10\xd2\x0f\x12\x0b\n\x06IMREAD\x10\xbe\x10\x12!\n\x1cPAIRWISE_EUCLIDEAN_DISTANCES\x10\x98\x11\x12!\n\x1cPAIRWISE_MANHATTAN_DISTANCES\x10\x99\x11\x12\x1e\n\x19PAIRWISE_COSINE_DISTANCES\x10\x9a\x11\x12!\n\x1cPAIRWISE_HAVERSINE_DISTANCES\x10\x9b\x11\x12\x12\n\rKD_TREE_TRAIN\x10\xb6\x11\x12\x12\n\rKD_TREE_QUERY\x10\xb7\x11\x12\x14\n\x0f\x42\x41LL_TREE_TRAIN\x10\xb8\x11\x12\x14\n\x0f\x42\x41LL_TREE_QUERY\x10\xb9\x11\x12\x16\n\x11\x46\x41ISS_BUILD_INDEX\x10\xba\x11\x12\x1e\n\x19\x46\x41ISS_TRAIN_SAMPLED_INDEX\x10\xbb\x11\x12\x10\n\x0b\x46\x41ISS_QUERY\x10\xbc\x11\x12\x15\n\x10KNEIGHBORS_GRAPH\x10\xbd\x11\x12\x1a\n\x15KMEANS_PLUS_PLUS_INIT\x10\xca\x11\x12#\n\x1eKMEANS_SCALABLE_PLUS_PLUS_INIT\x10\xcb\x11\x12\x1d\n\x18KMEANS_ELKAN_INIT_BOUNDS\x10\xcc\x11\x12\x18\n\x13KMEANS_ELKAN_UPDATE\x10\xcd\x11\x12\x1d\n\x18KMEANS_ELKAN_POSTPROCESS\x10\xce\x11\x12\x18\n\x13KMEANS_LLOYD_UPDATE\x10\xcf\x11\x12\x1d\n\x18KMEANS_LLOYD_POSTPROCESS\x10\xd0\x11\x12\x13\n\x0eKMEANS_INERTIA\x10\xd1\x11\x12$\n\x1fKMEANS_RELOCASTE_EMPTY_CLUSTERS\x10\xd2\x11\x12\x12\n\rXGBOOST_TRAIN\x10\xb9\x17\x12\x14\n\x0fXGBOOST_PREDICT\x10\xba\x17\x12\x0f\n\nTO_DMATRIX\x10\xbb\x17\x12\x12\n\rSTART_TRACKER\x10\xbc\x17\x12\x0f\n\nLGBM_TRAIN\x10\xcc\x17\x12\x11\n\x0cLGBM_PREDICT\x10\xcd\x17\x12\x0f\n\nLGBM_ALIGN\x10\xce\x17\x12\x13\n\x0eRUN_TENSORFLOW\x10\xc2\x17\x12\x10\n\x0bRUN_PYTORCH\x10\xc3\x17\x12\x17\n\x12\x43HECK_NON_NEGATIVE\x10\xe4\x19\x12\x12\n\rCHECK_TARGETS\x10\xe5\x19\x12\x16\n\x11\x41SSERT_ALL_FINITE\x10\xe6\x19\x12\x12\n\rIS_MULTILABEL\x10\xe7\x19\x12\x13\n\x0eTYPE_OF_TARGET\x10\xe8\x19\x12\x13\n\x0e\x41\x43\x43URACY_SCORE\x10\xe9\x19\x12\x15\n\x10REMOTE_FUNCATION\x10\x89\'\x12\x0f\n\nRUN_SCRIPT\x10\x8a\'\x12\x1f\n\x1aTENSOR_FROM_VINEYARD_CHUNK\x10\xa1\x1f\x12\x19\n\x14TENSOR_FROM_VINEYARD\x10\xa2\x1f\x12 \n\x1bTENSOR_STORE_VINEYARD_CHUNK\x10\xa3\x1f\x12$\n\x1fTENSOR_STORE_VINEYARD_CHUNK_MAP\x10\xa4\x1f\x12&\n!TENSOR_STORE_VINEYARD_GLOBAL_META\x10\xa5\x1f\x12\"\n\x1d\x44\x41TAFRAME_FROM_VINEYARD_CHUNK\x10\xa6\x1f\x12\x1c\n\x17\x44\x41TAFRAME_FROM_VINEYARD\x10\xa7\x1f\x12#\n\x1e\x44\x41TAFRAME_STORE_VINEYARD_CHUNK\x10\xa8\x1f\x12\'\n\"DATAFRAME_STORE_VINEYARD_CHUNK_MAP\x10\xa9\x1f\x12)\n$DATAFRAME_STORE_VINEYARD_GLOBAL_META\x10\xaa\x1f\x12\x13\n\rCHOLESKY_FUSE\x10\xb4\x84=\x12\x13\n\rFETCH_SHUFFLE\x10\xbe\x84=\x12\x0b\n\x05\x46\x45TCH\x10\xbf\x84=b\x06proto3')
  ,
  dependencies=[mars_dot_serialize_dot_protos_dot_value__pb2.DESCRIPTOR,])



_OPERANDDEF_OPERANDTYPE = _descriptor.EnumDescriptor(
  name='OperandType',
  full_name='OperandDef.OperandType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NULL', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SCALAR', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_DATA_SOURCE', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_ONES', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_ONES_LIKE', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_ZEROS', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_ZEROS_LIKE', index=6, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_EMPTY', index=7, number=7,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_EMPTY_LIKE', index=8, number=8,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_FULL', index=9, number=9,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_FULL_LIKE', index=10, number=25,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_ARANGE', index=11, number=10,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_INDICES', index=12, number=11,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_DIAG', index=13, number=12,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_EYE', index=14, number=13,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_LINSPACE', index=15, number=14,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_TRIU', index=16, number=15,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_TRIL', index=17, number=16,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_FROM_TILEDB', index=18, number=18,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_STORE_TILEDB', index=19, number=19,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_STORE_TILEDB_CONSOLIDATE', index=20, number=20,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_FROM_DATAFRAME', index=21, number=22,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_FROM_HDF5', index=22, number=27,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_STORE_HDF5', index=23, number=28,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_FROM_ZARR', index=24, number=29,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_STORE_ZARR', index=25, number=32,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATAFRAME_DATA_SOURCE', index=26, number=17,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATAFRAME_FROM_TENSOR', index=27, number=21,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SERIES_FROM_TENSOR', index=28, number=26,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATAFRAME_FROM_RECORDS', index=29, number=24,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SERIES_DATA_SOURCE', index=30, number=23,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INDEX_DATA_SOURCE', index=31, number=33,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATE_RANGE', index=32, number=34,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TIMEDELTA_RANGE', index=33, number=35,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MEMORY_USAGE', index=34, number=36,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TO_GPU', index=35, number=30,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TO_CPU', index=36, number=31,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RAND', index=37, number=41,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RANDN', index=38, number=42,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RANDINT', index=39, number=43,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RANDOM_INTEGERS', index=40, number=44,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RANDOM_SAMPLE', index=41, number=45,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RANDOM', index=42, number=46,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RANF', index=43, number=47,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_SAMPLE', index=44, number=48,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_BYTES', index=45, number=49,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_BETA', index=46, number=50,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_BINOMIAL', index=47, number=51,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_CHISQUARE', index=48, number=52,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_CHOICE', index=49, number=53,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_DIRICHLET', index=50, number=54,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_EXPONENTIAL', index=51, number=55,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_F', index=52, number=56,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_GAMMA', index=53, number=57,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_GEOMETRIC', index=54, number=58,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_GUMBEL', index=55, number=59,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_HYPERGEOMETRIC', index=56, number=60,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_LAPLACE', index=57, number=61,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_LOGISTIC', index=58, number=62,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_LOGNORMAL', index=59, number=63,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_LOGSERIES', index=60, number=64,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_MULTINOMIAL', index=61, number=65,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_MULTIVARIATE_NORMAL', index=62, number=66,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_NEGATIVE_BINOMIAL', index=63, number=67,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_NONCENTRAL_CHISQURE', index=64, number=68,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_NONCENTRAL_F', index=65, number=69,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_NORMAL', index=66, number=70,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_PARETO', index=67, number=71,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_PERMUTATION', index=68, number=72,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_POSSION', index=69, number=73,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_POWER', index=70, number=74,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_RAYLEIGH', index=71, number=75,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_SHUFFLE', index=72, number=76,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_STANDARD_CAUCHY', index=73, number=77,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_STANDARD_EXPONENTIAL', index=74, number=78,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_STANDARD_GAMMMA', index=75, number=79,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_STANDARD_NORMAL', index=76, number=80,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_STANDARD_T', index=77, number=81,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_TOMAXINT', index=78, number=82,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_TRIANGULAR', index=79, number=83,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_UNIFORM', index=80, number=84,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_VONMISES', index=81, number=85,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_WALD', index=82, number=86,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_WEIBULL', index=83, number=87,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAND_ZIPF', index=84, number=88,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PERMUTATION', index=85, number=89,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNIQUE', index=86, number=90,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ADD', index=87, number=101,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SUB', index=88, number=102,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MUL', index=89, number=103,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DIV', index=90, number=104,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TRUEDIV', index=91, number=105,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FLOORDIV', index=92, number=106,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='POW', index=93, number=107,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MOD', index=94, number=108,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FMOD', index=95, number=109,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOGADDEXP', index=96, number=110,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOGADDEXP2', index=97, number=111,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NEGATIVE', index=98, number=112,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='POSITIVE', index=99, number=113,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ABSOLUTE', index=100, number=114,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FABS', index=101, number=115,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ABS', index=102, number=116,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RINT', index=103, number=117,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SIGN', index=104, number=118,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONJ', index=105, number=119,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EXP', index=106, number=120,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EXP2', index=107, number=121,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOG', index=108, number=122,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOG2', index=109, number=123,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOG10', index=110, number=124,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EXPM1', index=111, number=125,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOG1P', index=112, number=126,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SQRT', index=113, number=127,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SQUARE', index=114, number=128,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CBRT', index=115, number=129,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RECIPROCAL', index=116, number=130,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EQ', index=117, number=131,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NE', index=118, number=132,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LT', index=119, number=133,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LE', index=120, number=134,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GT', index=121, number=135,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GE', index=122, number=136,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SIN', index=123, number=137,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COS', index=124, number=138,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TAN', index=125, number=139,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARCSIN', index=126, number=140,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARCCOS', index=127, number=141,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARCTAN', index=128, number=142,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARCTAN2', index=129, number=143,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HYPOT', index=130, number=144,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SINH', index=131, number=145,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COSH', index=132, number=146,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TANH', index=133, number=147,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARCSINH', index=134, number=148,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARCCOSH', index=135, number=149,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARCTANH', index=136, number=150,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DEG2RAD', index=137, number=151,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAD2DEG', index=138, number=152,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BITAND', index=139, number=153,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BITOR', index=140, number=154,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BITXOR', index=141, number=155,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INVERT', index=142, number=156,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LSHIFT', index=143, number=157,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RSHIFT', index=144, number=158,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AND', index=145, number=159,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OR', index=146, number=160,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='XOR', index=147, number=161,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NOT', index=148, number=162,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MAXIMUM', index=149, number=163,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MINIMUM', index=150, number=164,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AROUND', index=151, number=165,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FLOAT_POWER', index=152, number=166,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FMAX', index=153, number=167,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FMIN', index=154, number=168,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ISFINITE', index=155, number=169,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ISINF', index=156, number=170,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ISNAN', index=157, number=171,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SIGNBIT', index=158, number=172,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COPYSIGN', index=159, number=173,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NEXTAFTER', index=160, number=174,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SPACING', index=161, number=175,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LDEXP', index=162, number=176,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FREXP', index=163, number=177,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MODF', index=164, number=178,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FLOOR', index=165, number=179,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CEIL', index=166, number=180,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TRUNC', index=167, number=181,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DEGREES', index=168, number=182,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RADIANS', index=169, number=183,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CLIP', index=170, number=184,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ISREAL', index=171, number=185,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ISCOMPLEX', index=172, number=186,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REAL', index=173, number=187,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IMAG', index=174, number=188,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FIX', index=175, number=189,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='I0', index=176, number=190,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SINC', index=177, number=191,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NAN_TO_NUM', index=178, number=192,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ISCLOSE', index=179, number=193,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DIVMOD', index=180, number=194,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ANGLE', index=181, number=195,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SET_REAL', index=182, number=196,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SET_IMAG', index=183, number=197,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GAMMALN', index=184, number=200,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ERF', index=185, number=201,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ENTR', index=186, number=202,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REL_ENTR', index=187, number=203,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PDIST', index=188, number=231,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CDIST', index=189, number=232,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SQUAREFORM', index=190, number=233,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TREE_ADD', index=191, number=251,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TREE_MULTIPLY', index=192, number=252,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUMSUM', index=193, number=301,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUMPROD', index=194, number=302,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PROD', index=195, number=303,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SUM', index=196, number=304,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MAX', index=197, number=305,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MIN', index=198, number=306,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ALL', index=199, number=307,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ANY', index=200, number=308,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MEAN', index=201, number=309,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARGMAX', index=202, number=310,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARGMIN', index=203, number=311,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANSUM', index=204, number=312,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANMAX', index=205, number=313,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANMIN', index=206, number=314,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANPROD', index=207, number=315,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANMEAN', index=208, number=316,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANARGMAX', index=209, number=317,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANARGMIN', index=210, number=318,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COUNT_NONZERO', index=211, number=319,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MOMENT', index=212, number=320,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANMOMENT', index=213, number=321,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='VAR', index=214, number=322,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STD', index=215, number=323,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANVAR', index=216, number=324,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANSTD', index=217, number=325,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANCUMSUM', index=218, number=326,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NANCUMPROD', index=219, number=327,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COUNT', index=220, number=343,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUMMAX', index=221, number=344,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUMMIN', index=222, number=345,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUMCOUNT', index=223, number=346,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RESHAPE', index=224, number=401,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SLICE', index=225, number=402,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INDEX', index=226, number=403,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INDEXSETVALUE', index=227, number=404,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONCATENATE', index=228, number=405,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RECHUNK', index=229, number=406,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ASTYPE', index=230, number=407,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TRANSPOSE', index=231, number=408,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SWAPAXES', index=232, number=409,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BROADCAST_TO', index=233, number=410,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STACK', index=234, number=411,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='WHERE', index=235, number=412,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CHOOSE', index=236, number=413,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NONZERO', index=237, number=414,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARGWHERE', index=238, number=415,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNRAVEL_INDEX', index=239, number=416,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RAVEL_MULTI_INDEX', index=240, number=417,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARRAY_SPLIT', index=241, number=418,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SQUEEZE', index=242, number=419,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DIGITIZE', index=243, number=420,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REPEAT', index=244, number=421,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='COPYTO', index=245, number=422,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ISIN', index=246, number=423,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SEARCHSORTED', index=247, number=428,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SORT', index=248, number=429,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HISTOGRAM', index=249, number=430,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HISTOGRAM_BIN_EDGES', index=250, number=431,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PARTITION', index=251, number=432,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='QUANTILE', index=252, number=440,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FILL_DIAGONAL', index=253, number=441,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NORMALIZE', index=254, number=442,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TOPK', index=255, number=443,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TRAPZ', index=256, number=444,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GET_SHAPE', index=257, number=445,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FANCY_INDEX_DISTRIBUTE', index=258, number=424,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FANCY_INDEX_CONCAT', index=259, number=425,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSORDOT', index=260, number=501,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DOT', index=261, number=502,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MATMUL', index=262, number=503,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CHOLESKY', index=263, number=510,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='QR', index=264, number=511,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SVD', index=265, number=512,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LU', index=266, number=513,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SOLVE_TRIANGULAR', index=267, number=520,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INV', index=268, number=521,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NORM', index=269, number=530,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FFT', index=270, number=601,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IFFT', index=271, number=602,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FFT2', index=272, number=603,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IFFT2', index=273, number=604,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FFTN', index=274, number=605,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IFFTN', index=275, number=606,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RFFT', index=276, number=607,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IRFFT', index=277, number=608,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RFFT2', index=278, number=609,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IRFFT2', index=279, number=610,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RFFTN', index=280, number=611,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IRFFTN', index=281, number=612,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HFFT', index=282, number=613,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IHFFT', index=283, number=614,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FFTFREQ', index=284, number=615,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FFTFREQ_CHUNK', index=285, number=616,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RFFTFREQ', index=286, number=617,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FFTSHIFT', index=287, number=618,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IFFTSHIFT', index=288, number=619,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EINSUM', index=289, number=630,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SPARSE_MATRIX_DATA_SOURCE', index=290, number=701,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DENSE_TO_SPARSE', index=291, number=702,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SPARSE_TO_DENSE', index=292, number=703,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MAP', index=293, number=710,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DESCRIBE', index=294, number=712,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FILL_NA', index=295, number=713,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AGGREGATE', index=296, number=714,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STRING_METHOD', index=297, number=715,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATETIME_METHOD', index=298, number=716,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='APPLY', index=299, number=717,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TRANSFORM', index=300, number=718,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CHECK_NA', index=301, number=719,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DROP_NA', index=302, number=720,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NUNIQUE', index=303, number=721,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUT', index=304, number=722,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SHIFT', index=305, number=723,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATAFRAME_DIFF', index=306, number=724,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='VALUE_COUNTS', index=307, number=725,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TO_DATETIME', index=308, number=726,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATAFRAME_DROP', index=309, number=727,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DROP_DUPLICATES', index=310, number=728,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MELT', index=311, number=729,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SET_LABEL', index=312, number=730,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RENAME', index=313, number=731,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INSERT', index=314, number=732,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FUSE', index=315, number=801,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ENTER', index=316, number=901,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LEAVE', index=317, number=902,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FIX_LATEST', index=318, number=903,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IF_ELSE', index=319, number=904,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NEXT_ITER', index=320, number=905,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TABLE_COO', index=321, number=1003,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STORE_COO', index=322, number=1004,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SHUFFLE_PROXY', index=323, number=2001,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATAFRAME_INDEX_ALIGN', index=324, number=2004,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNIQUE_INVERSE_REDUCE', index=325, number=2013,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATAFRAME_SET_INDEX', index=326, number=2020,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATAFRAME_ILOC_GETITEM', index=327, number=2021,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATAFRAME_ILOC_SETITEM', index=328, number=2022,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATAFRAME_LOC_GETITEM', index=329, number=2023,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATAFRAME_LOC_SETITEM', index=330, number=2024,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATAFRAME_SHUFFLE_MERGE', index=331, number=2010,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATAFRAME_SHUFFLE_MERGE_ALIGN', index=332, number=2011,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='APPEND', index=333, number=2015,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RESET_INDEX', index=334, number=2028,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REINDEX', index=335, number=2029,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GROUPBY', index=336, number=2030,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GROUPBY_AGG', index=337, number=2033,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GROUPBY_CONCAT', index=338, number=2034,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PSRS_SORT_REGULAR_SMAPLE', index=339, number=2040,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PSRS_CONCAT_PIVOT', index=340, number=2041,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PSRS_SHUFFLE', index=341, number=2042,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PSRS_ALIGN', index=342, number=2043,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CALC_PARTITIONS_INFO', index=343, number=2046,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PARTITION_MERGED', index=344, number=2047,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SORT_VALUES', index=345, number=2050,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SORT_INDEX', index=346, number=2051,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ROLLING_AGG', index=347, number=2060,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EXPANDING_AGG', index=348, number=2061,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EWM_AGG', index=349, number=2062,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='READ_CSV', index=350, number=2100,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TO_CSV', index=351, number=2101,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='READ_PARQUET', index=352, number=2103,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TO_PARQUET', index=353, number=2104,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='READ_SQL', index=354, number=2105,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TO_SQL', index=355, number=2108,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TO_CSV_STAT', index=356, number=2102,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STANDARDIZE_RANGE_INDEX', index=357, number=2107,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SUCCESSORS_EXCLUSIVE', index=358, number=2002,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IMREAD', index=359, number=2110,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PAIRWISE_EUCLIDEAN_DISTANCES', index=360, number=2200,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PAIRWISE_MANHATTAN_DISTANCES', index=361, number=2201,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PAIRWISE_COSINE_DISTANCES', index=362, number=2202,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PAIRWISE_HAVERSINE_DISTANCES', index=363, number=2203,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='KD_TREE_TRAIN', index=364, number=2230,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='KD_TREE_QUERY', index=365, number=2231,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BALL_TREE_TRAIN', index=366, number=2232,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BALL_TREE_QUERY', index=367, number=2233,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FAISS_BUILD_INDEX', index=368, number=2234,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FAISS_TRAIN_SAMPLED_INDEX', index=369, number=2235,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FAISS_QUERY', index=370, number=2236,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='KNEIGHBORS_GRAPH', index=371, number=2237,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='KMEANS_PLUS_PLUS_INIT', index=372, number=2250,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='KMEANS_SCALABLE_PLUS_PLUS_INIT', index=373, number=2251,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='KMEANS_ELKAN_INIT_BOUNDS', index=374, number=2252,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='KMEANS_ELKAN_UPDATE', index=375, number=2253,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='KMEANS_ELKAN_POSTPROCESS', index=376, number=2254,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='KMEANS_LLOYD_UPDATE', index=377, number=2255,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='KMEANS_LLOYD_POSTPROCESS', index=378, number=2256,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='KMEANS_INERTIA', index=379, number=2257,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='KMEANS_RELOCASTE_EMPTY_CLUSTERS', index=380, number=2258,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='XGBOOST_TRAIN', index=381, number=3001,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='XGBOOST_PREDICT', index=382, number=3002,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TO_DMATRIX', index=383, number=3003,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='START_TRACKER', index=384, number=3004,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LGBM_TRAIN', index=385, number=3020,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LGBM_PREDICT', index=386, number=3021,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LGBM_ALIGN', index=387, number=3022,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RUN_TENSORFLOW', index=388, number=3010,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RUN_PYTORCH', index=389, number=3011,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CHECK_NON_NEGATIVE', index=390, number=3300,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CHECK_TARGETS', index=391, number=3301,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ASSERT_ALL_FINITE', index=392, number=3302,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IS_MULTILABEL', index=393, number=3303,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TYPE_OF_TARGET', index=394, number=3304,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ACCURACY_SCORE', index=395, number=3305,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='REMOTE_FUNCATION', index=396, number=5001,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RUN_SCRIPT', index=397, number=5002,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_FROM_VINEYARD_CHUNK', index=398, number=4001,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_FROM_VINEYARD', index=399, number=4002,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_STORE_VINEYARD_CHUNK', index=400, number=4003,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_STORE_VINEYARD_CHUNK_MAP', index=401, number=4004,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TENSOR_STORE_VINEYARD_GLOBAL_META', index=402, number=4005,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATAFRAME_FROM_VINEYARD_CHUNK', index=403, number=4006,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATAFRAME_FROM_VINEYARD', index=404, number=4007,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATAFRAME_STORE_VINEYARD_CHUNK', index=405, number=4008,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATAFRAME_STORE_VINEYARD_CHUNK_MAP', index=406, number=4009,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATAFRAME_STORE_VINEYARD_GLOBAL_META', index=407, number=4010,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CHOLESKY_FUSE', index=408, number=999988,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FETCH_SHUFFLE', index=409, number=999998,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FETCH', index=410, number=999999,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=206,
  serialized_end=7290,
)
_sym_db.RegisterEnumDescriptor(_OPERANDDEF_OPERANDTYPE)


_OPERANDDEF_ATTRENTRY = _descriptor.Descriptor(
  name='AttrEntry',
  full_name='OperandDef.AttrEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='OperandDef.AttrEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='OperandDef.AttrEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=152,
  serialized_end=203,
)

_OPERANDDEF = _descriptor.Descriptor(
  name='OperandDef',
  full_name='OperandDef',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='OperandDef.type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='id', full_name='OperandDef.id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='attr', full_name='OperandDef.attr', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_OPERANDDEF_ATTRENTRY, ],
  enum_types=[
    _OPERANDDEF_OPERANDTYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=75,
  serialized_end=7290,
)

_OPERANDDEF_ATTRENTRY.fields_by_name['value'].message_type = mars_dot_serialize_dot_protos_dot_value__pb2._VALUE
_OPERANDDEF_ATTRENTRY.containing_type = _OPERANDDEF
_OPERANDDEF.fields_by_name['attr'].message_type = _OPERANDDEF_ATTRENTRY
_OPERANDDEF_OPERANDTYPE.containing_type = _OPERANDDEF
DESCRIPTOR.message_types_by_name['OperandDef'] = _OPERANDDEF
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

OperandDef = _reflection.GeneratedProtocolMessageType('OperandDef', (_message.Message,), dict(

  AttrEntry = _reflection.GeneratedProtocolMessageType('AttrEntry', (_message.Message,), dict(
    DESCRIPTOR = _OPERANDDEF_ATTRENTRY,
    __module__ = 'mars.serialize.protos.operand_pb2'
    # @@protoc_insertion_point(class_scope:OperandDef.AttrEntry)
    ))
  ,
  DESCRIPTOR = _OPERANDDEF,
  __module__ = 'mars.serialize.protos.operand_pb2'
  # @@protoc_insertion_point(class_scope:OperandDef)
  ))
_sym_db.RegisterMessage(OperandDef)
_sym_db.RegisterMessage(OperandDef.AttrEntry)


_OPERANDDEF_ATTRENTRY._options = None
# @@protoc_insertion_point(module_scope)
